/****************************************************************************

 ■SleepLamp
		Copyright(C) Denshi Usagi

****************************************************************************/

// PIC12F1822 Configuration Bit Settings

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = SWDTEN    // Watchdog Timer Enable (WDT controlled by the SWDTEN bit in the WDTCON register)
#pragma config PWRTE = ON       // Power-up Timer Enable (PWRT enabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Memory Code Protection (Data memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = ON        // Internal/External Switchover (Internal/External Switchover mode is enabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PLLEN = OFF      // PLL Enable (4x PLL disabled)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

#include <xc.h>

typedef unsigned char U8;
typedef unsigned short U16;
typedef U8 BOOL;

#define TRUE	1
#define FALSE	0

//=================================================================

#define NIGHT_THR		1000
#define LIGHT_LV		525
//#define LIGHT_PS		0b01101 // 1:262144 (218) (Interval 8s typ)
//#define LIGHT_PS		0b01110	// 1:524288 (219) (Interval 16s typ)
#define LIGHT_PS		0b01111 // 1:1048576 (220) (Interval 32s typ)

#if 0 // Test
#undef NIGHT_THR
#undef LIGHT_LV
#undef LIGHT_PS
#define NIGHT_THR		1000
#define LIGHT_LV		1000
#define LIGHT_PS		0b01010 // 1:32768 (Interval 1s typ)
#endif

#define DET_DELTA		25
#define DAY_THR			(NIGHT_THR + 4)

//#define NOUSE_WDT
#ifndef NOUSE_WDT
#define Sleep()			SLEEP()
#else
#define Sleep()			DelayMs(1000)
#endif

//=================================================================

#define LED_STEP		255
const U16 s_PwmTbl[LED_STEP] = { 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 15, 15, 16, 16, 17, 17, 17, 18, 18, 19, 20, 20, 21, 21, 22, 22, 23, 23, 24, 25, 25, 26, 27, 27, 28, 29, 29, 30, 31, 32, 32, 33, 34, 35, 36, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 56, 57, 58, 59, 61, 62, 63, 65, 66, 67, 69, 70, 72, 73, 75, 76, 78, 80, 81, 83, 85, 87, 88, 90, 92, 94, 96, 98, 100, 102, 104, 106, 108, 110, 113, 115, 117, 120, 122, 125, 127, 130, 132, 135, 138, 140, 143, 146, 149, 152, 155, 158, 161, 164, 167, 171, 174, 178, 181, 185, 188, 192, 196, 200, 204, 208, 212, 216, 220, 224, 229, 233, 238, 242, 247, 252, 257, 262, 267, 272, 277, 283, 288, 294, 299, 305, 311, 317, 323, 329, 336, 342, 349, 356, 362, 369, 376, 384, 391, 399, 406, 414, 422, 430, 438, 447, 455, 464, 473, 482, 491, 500, 510, 520, 530, 540, 550, 561, 571, 582, 593, 605, 616, 628, 640, 652, 664, 677, 690, 703, 716, 730, 744, 758, 773, 787, 802, 817, 833, 849, 865, 881, 898, 915, 932, 950, 968, 986, 1005, 1023 };

static BOOL s_bFast;
static U16 s_uDelay;
static U8 s_nTmr0;
static BOOL s_bDir;
static U8 s_uStep;

static U16 s_uPreCds = 0;
static U16 s_uCurCds = 0;

//=================================================================

static void InitSystem(void);
static void Run(void);

static void GetLight(void);
static void Lighting(U8 uWdtPs);
static void LedCtrl(BOOL bOn, U8 uSpeed);
static void PowerOn(BOOL bOn);
static void SleepDelay(U8 uWdtPs);
static void EnableWDT(BOOL bEnable, U8 nPreScale);
static void SetClock(BOOL bFast);
static void DelayMs(U16 uMs);

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();

	PowerOn(TRUE);
	for (int i = 0; i < 3; i++)
	{
		LATA2 = 1; DelayMs(300);
		LATA2 = 0; DelayMs(300);
	}
	PowerOn(FALSE);
	
	Run();
}

// ●システム初期化
static void InitSystem(void)
{
	// ポート初期化
	PORTA = LATA = 0x00;
	
	// RA5(Power)
	TRISAbits.TRISA5 = 1; // Hi-Z
	
	// RA2(PWM)
	TRISAbits.TRISA2 = 0;
	ANSELAbits.ANSA2 = 0;
	
	// ADC(CDS)
	ADCON1 = 0b10110000; // Right justified, FRC, VREF+ is connected to AVDD
	ADCON0 = 0b00001101; // AN3, ADC is enabled

	// PWM
	CCP1SEL = 0; // CCP1/P1A function is on RA2

	// Timer0(Sweep, Delay)
	TMR0CS = 0;
	OPTION_REGbits.PS = 0b101; // 1 : 64

	// Timer2(PWM)
	T2CON = 0b00000000; // 1:1 Postscaler, Prescaler is 1
	PR2 = 255;

	// 割込み許可
	PEIE = 1;
	GIE = 1;

	// クロック(500KHz⇒31KHz)
	SetClock(FALSE);
}

// ●実行
static void Run(void)
{
	//for(;;)GetLight();

	for(;;)
	{
		// デイモード
		for(;;)
		{
			GetLight();
			if (s_uCurCds > NIGHT_THR)
			{
				SleepDelay(0b01010); // 1:32768 (Interval 1s typ)
				continue;
			}
			break;
		}

		// ナイトモード
		for(;;)
		{
			if ( (s_uCurCds <= LIGHT_LV) && (s_uPreCds > s_uCurCds) && ((s_uPreCds - s_uCurCds) >= DET_DELTA) )
			{
				Lighting(LIGHT_PS);
				DelayMs(500); // 電源安定化待ち

				s_uCurCds = 0;
			}
			else
			{
				//SleepDelay(0b00111); // 1:4096 (Interval 128 ms typ)
				SleepDelay(0b01000); // 1:8192 (Interval 256 ms typ)
			}

			GetLight();
			if (s_uCurCds >= DAY_THR)
				break;
		}
	}
}

// ●光量検出
static void GetLight(void)
{
	GO_nDONE = 1;
	while(GO_nDONE);

	s_uPreCds = s_uCurCds;
	s_uCurCds = (ADRESH << 8) | ADRESL;
}

// ●照明
static void Lighting(U8 uWdtPs)
{
	SetClock(TRUE);

	LedCtrl(TRUE, 15);
	SleepDelay(uWdtPs);
	LedCtrl(FALSE, 180);

	SetClock(FALSE);
}

// ●LED制御
static void LedCtrl(BOOL bOn, U8 uSpeed)
{
	if (bOn)
	{
		TMR2 = 0;
		TMR2ON = 1;

		CCPR1L = 0;
		CCP1CON = 0b00001100; // Single output, PWM mode: P1A active-high
		PowerOn(TRUE);
	}

	s_bDir = bOn;
	s_uStep = (s_bDir)? 0 : (LED_STEP - 1);
	s_nTmr0 = 0xFF - uSpeed + 1;

	PSA = 0; // Prescaler is assigned to the Timer0 module
	TMR0 = s_nTmr0;
	TMR0IF = 0;
	TMR0IE = 1;

	while (TMR0IE);

	if (!bOn)
	{
		PowerOn(FALSE);
		CCP1CON = 0;

		TMR2ON = 0;
	}
}

// ●DC-DC(LED電源)制御
static void PowerOn(BOOL bOn)
{
	if (bOn)
	{
		TRISAbits.TRISA5 = 0; // Digital output L
	}
	else
	{
		TRISAbits.TRISA5 = 1; // Hight Z
	}
}

// ●スリープ
static void SleepDelay(U8 uWdtPs)
{
	EnableWDT(TRUE, uWdtPs);
	CLRWDT();
	Sleep();
	EnableWDT(FALSE, 0);
}

// ●WDT制御
static void EnableWDT(BOOL bEnable, U8 nPreScale)
{
	if (bEnable)
	{
		WDTCONbits.WDTPS = nPreScale;
#ifndef NOUSE_WDT
		SWDTEN = 1;
#endif
	}
	else {
		SWDTEN = 0;
	}
}

// ●クロック設定
static void SetClock(BOOL bFast)
{
	if (bFast)
	{
		// 4MHz HF
		OSCCONbits.IRCF = 0b1101;
	}
	else
	{
		// 31 kHz LF
		OSCCONbits.IRCF = 0b0000;
	}
	s_bFast = bFast;
}

// ●ウェイト
static void DelayMs(U16 uMs)
{
	if (s_bFast)
	{
		s_nTmr0 = 0xFF - 250 + 1; // 0.25ms at 4MHz Fosc
		s_uDelay = uMs * 4;
	}
	else
	{
		s_nTmr0 = 0xFF - /*80*/68 + 1; // 10ms at 31KHz Fosc
		s_uDelay = uMs / 10;
	}
	
	PSA = 1; // Prescaler is not assigned to the Timer0 module
	TMR0 = s_nTmr0;
	TMR0IF = 0;
	TMR0IE = 1;

	while (TMR0IE);
}

// ●割込み
void __interrupt() Interrupt(void)
{
	if (TMR0IF)
	{
		// Timer0
		TMR0 += s_nTmr0;
		if (s_uDelay == 0)
		{
			// Sweep
			if (s_bDir)
			{
				if (++s_uStep == (LED_STEP - 1))
					TMR0IE = 0;
			}
			else
			{
				if (--s_uStep == 0)
					TMR0IE = 0;
			}

			U16 nVal = s_PwmTbl[s_uStep];
			CCPR1L = (U8)(nVal >> 2);
			CCP1CONbits.DC1B = nVal & 3;
		}
		else
		{
			// Delay
			if (--s_uDelay == 0)
				TMR0IE = 0;
		}
		TMR0IF = 0;
	}
}
